#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sha2.h>

const size_t gsha_keylen = 64;
const size_t gsha_hashlen = 64;

struct gsha {
	uint64_t start[8];
	uint64_t regs[8];
	uint32_t srcidx[64];
	uint32_t rolls[64];
	uint32_t dstidx[64];
};


static const uint32_t rollamts[8] = {
	3, 8, 21, 31, 39, 48, 57, 60
};

struct gsha *
gsha_new(uint8_t *key)
{
	struct gsha *gsha;

	if (!(gsha = malloc(sizeof(*gsha))))
		return NULL;
	memcpy(gsha->start, key, gsha_hashlen);
	for (int i = 0; i < gsha_keylen; i++) {
		uint8_t opcode = key[i];

		int srcreg = opcode & 0x7;
		opcode >>= 3;
		int rollidx = opcode & 0x7;
		opcode >>= 3;
		int dstreg = srcreg + 2 + (opcode & 0x3) & 0x7;
		opcode >>= 2;

		gsha->srcidx[i] = srcreg;
		gsha->rolls[i] = rollamts[rollidx];
		gsha->dstidx[i] = dstreg;
	}
	return gsha;
}

void
gsha_free(struct gsha *gsha)
{
	free(gsha);
}

static uint64_t
roll(uint64_t x, uint32_t r)
{
	uint64_t y;
	y = (x >> r) | (x << 64 - r);
	return y;
}

static void
mix(struct gsha *gsha, uint64_t *input)
{
	for (int i = 0; i < 64; i++) {
		uint64_t srcval = gsha->regs[gsha->srcidx[i]];
		uint64_t dstval = input[i & 7] ^ roll(srcval, gsha->rolls[i]);
		gsha->regs[gsha->dstidx[i]] += dstval;
	}
}

void
gsha_hash(struct gsha *gsha, void *data, size_t datalen, void *hash)
{
	uint8_t *dp = data;
	union {
		uint64_t input[8];
		unsigned char c[gsha_hashlen];
	} x;

	memcpy(gsha->regs, gsha->start, sizeof(gsha->regs));
	while (datalen > gsha_hashlen) {
		memcpy(x.input, dp, gsha_hashlen);
		mix(gsha, x.input);
		datalen -= gsha_hashlen;
		dp += gsha_hashlen;
	}
	for (int i = 0; i < gsha_hashlen; i++)
		x.c[i] = gsha_hashlen - i;
	memcpy(x.input, dp, datalen);
	mix(gsha, x.input);
	memcpy(hash, gsha->regs, gsha_hashlen);
}

struct gsha *
gsha_frompass(char *pass, size_t passlen, uint8_t *salt, size_t saltlen)
{
	uint8_t key[gsha_keylen];
	SHA2_CTX ctx;

	SHA512Init(&ctx);
	SHA512Update(&ctx, pass, passlen);
	SHA512Update(&ctx, salt, saltlen);
	SHA512Final(key, &ctx);

	return gsha_new(key);
}

int
gsha_kdf(char *pass, size_t passlen, uint8_t *salt, size_t saltlen,
    int rounds, uint8_t *key, size_t keylen)
{
	struct gsha *gsha;
	size_t amt, inputlen, saltoffset;
	uint8_t *input;
	uint32_t counter = 0;

	if (!(gsha = gsha_frompass(pass, passlen, salt, saltlen)))
		return -1;
	if (!(input = malloc(4 + passlen + saltlen + gsha_hashlen))) {
		gsha_free(gsha);
		return -1;
	}
	saltoffset = 4 + passlen;
	while (keylen > 0) {
		inputlen = saltoffset + saltlen;
		*(uint32_t *)input = counter++;
		memcpy(input + 4, pass, passlen);
		memcpy(input + saltoffset, salt, saltlen);
		for (int i = 0; i < rounds; i++) {
			gsha_hash(gsha, input, inputlen, input + saltoffset);
			inputlen = saltoffset + gsha_hashlen;
		}
		amt = keylen < gsha_hashlen ? keylen : gsha_hashlen;
		memcpy(key, input + saltoffset, amt);
		keylen -= amt;
		key += amt;
	}
	gsha_free(gsha);
	return 0;
}


int
main(int argc, char **argv)
{
	uint8_t key[32];
	
	gsha_kdf(argv[1], strlen(argv[1]), argv[2], strlen(argv[2]), atoi(argv[3]),
	    key, sizeof(key));
	for (int i = 0; i < 32; i++)
		printf("%2.2x", key[i]);
	printf("\n");
}
